//
// Created by huli on 2020/7/20.
//

#ifndef NNXX_TESTS_CARRIER_BASE_H
#define NNXX_TESTS_CARRIER_BASE_H

#include "../error_code/error_code.h"
#include <glog/logging.h>
#include <thread>
#include <mutex>
#include "../tool/thread_condition.h"
#include "../dispatch/carrier_task.h"

//搬运器的基类, 楚天项目就是电梯和中跑车的合并
class Carrier_base
{
public:
	//搬运器底层通信延时5000ms
#define	CARRIER_COMMUNICATION_OVER_TIME_MS	5000

	//搬运器状态,
	enum Carrier_status
	{
		E_UNKNOW               	= 0,    //未知
		E_READY               	= 1,    //准备，待机
		E_BUSY					= 2, 	//工作正忙

		E_ONE_LEVEL_WORK		= 3,	//一级工作状态, 就是普通的移动任务
		E_ONE_LEVEL_OVER		= 4, 	//一级工作完成,
		E_TWO_LEVEL_WORK		= 5,	//二级工作状态, 就是紧急避让, 插队任务(会在执行一级任务的过程中, 插队并优先执行二级任务)
		E_TWO_LEVEL_OVER		= 6, 	//一级工作完成,
		E_THREE_LEVEL_WORK		= 7,	//三级工作任务, 就是锁定硬件资源, 不让插队, (除非急停或者取消任务)
		E_THREE_LEVEL_OVER		= 8, 	//一级工作完成,

		E_FAULT					= 100,	//故障
		E_DISCONNECT			= 101, 	//通信故障
	};

	//指令完成状态, 搬运器答复指令, 返回任务完成的情况
	enum Respons_status
	{
		RESPONS_NORMAL             	= 0,    //正常
		RESPONS_OVER               	= 1,    //任务完成
		RESPONS_MINOR_ERROR			= 100,	//一般故障, 可恢复
		RESPONS_CRITICAL_ERROR		= 101,	//致命故障,不可恢复
	};

public:
	Carrier_base();
	Carrier_base(const Carrier_base& other)= default;
	Carrier_base& operator =(const Carrier_base& other)= default;
	~Carrier_base();
public://API functions
	//搬运器 初始化
	virtual Error_manager carrier_base_init();
	//搬运器 反初始化
	virtual Error_manager carrier_base_uninit();

	//处理一级任务, 普通任务
	virtual Error_manager execute_one_level_task(std::shared_ptr<Carrier_task> p_carrier_task);
	//处理二级任务, 插队任务
	virtual Error_manager execute_two_level_task(std::shared_ptr<Carrier_task> p_carrier_task);
	//处理三级任务, 核心任务
	virtual Error_manager execute_three_level_task(std::shared_ptr<Carrier_task> p_carrier_task);
	//检查状态,是否正常运行
	Error_manager check_status();

	//注意了, 调度任务允许同时接受多个任务
	//判断能否执行一级任务
	Error_manager	check_one_level_task();
	//判断能否执行二级任务
	Error_manager	check_two_level_task();
	//判断能否执行三级任务
	Error_manager	check_three_level_task();

	//结束任务单，里面会根据任务的故障等级修正 任务单的状态
	Error_manager end_task(std::shared_ptr<Carrier_task> carrier_task);
	//取消任务单，由发送方提前取消任务单
	Error_manager cancel_task(std::shared_ptr<Carrier_task> carrier_task);

	//判断是否为待机，如果已经准备好，则可以执行任务。
	bool is_ready();
public://get or set member variable
	Carrier_status get_carrier_status();

protected://member functions
	//执行外界任务的执行函数
	void execute_thread_fun();

	//更新设备底层通信数据
	virtual Error_manager update_device_communication();

	//取消下发的指令
	virtual Error_manager cancel_command();
protected://member variable
	std::atomic<Carrier_status>			m_carrier_status;			//搬运器总状态, 控制任务流程
	int 								m_carrier_id;				//搬运器id, (楚天项目和单元号对应)
	
	//下发到硬件
	//搬运器的目标车位
	unsigned short 						m_parkspace_unit_id;		//车位单元号(楚天项目为1~3)
	unsigned short 						m_parkspace_floor_id;		//车位楼层号(楚天项目为1~14)
	unsigned short 						m_parkspace_room_id;		//车位同层的房间号	(楚天项目一楼为1~4, 楼上为1~6)
	std::string							m_command_key;				//任务唯一码, 用作识别
	unsigned char						m_motion_direction;			//调度方向, 根据停车取车选择出入口, 	1=入口，0=非入口
	float 								m_car_center_x;				//汽车的中心坐标x
	float 								m_car_center_y;				//汽车的中心坐标y
	float 								m_car_angle;				//汽车的旋转角(角度-90~90)

	//硬件反馈
	//搬运器当前坐标
	std::string							m_respons_key;				//应答的唯一码, 用作识别
	Respons_status						m_respons_status;			//指令完成状态, 搬运器答复指令, 返回任务完成的情况



	//硬件状态, 目前只做显示判断
	std::chrono::system_clock::time_point	m_status_updata_time;	//状态更新时间点
	unsigned char						m_last_heartbeat;			//上一次的心跳

	//硬件坐标,
	float 								m_carrier_coordinate_x;			//搬运器坐标x轴, 中跑车控制横向移动
	float 								m_carrier_coordinate_y;			//搬运器坐标y轴, (本模块默认为0, 由AGV模块去控制纵向移动)
	float 								m_carrier_coordinate_z;			//搬运器坐标z轴, 电梯控制上下移动
	float 								m_carrier_angle;				//搬运器角度
	//搬运器设备的状态(0bit为1:急停正常, 1bit为1:处于安全位置, 2bit为1:静止状态, 3bit为1:可执行新指令, 4bit为1:存在故障)
	unsigned char						m_carrier_device_status;	//搬运器设备的状态, 楚天项目就是中跑车
	unsigned char						m_carrier_error_code[50];	//搬运器设备的报警信息位
	std::string							m_carrier_error_description;//搬运器设备的报警状态描述
	//升降机设备的状态(0bit为1:急停正常, 1bit为1:处于安全位置, 2bit为1:静止状态, 3bit为1:可执行新指令, 4bit为1:存在故障)
	unsigned char						m_elevator_device_status;	//升降机设备状态, 楚天项目就是电梯
	unsigned char						m_elevator_error_code[50];	//升降机设备的报警信息位
	std::string							m_elevator_error_description;//升降机设备的报警状态描述


	//任务执行线程
	std::thread*        				mp_execute_thread;   			//执行的线程指针，内存由本类管理
	Thread_condition					m_execute_condition;			//执行的条件变量

	std::shared_ptr<Carrier_task>		mp_carrier_one_level_task;		//搬运器的普通任务, (一般移动和等待指令, 允许被打断)
	std::shared_ptr<Carrier_task>		mp_carrier_two_level_task;		//搬运器的优先任务, (紧急避让和插队指令, 允许打断普通任务)
	std::shared_ptr<Carrier_task>		mp_carrier_three_level_task;	//搬运器的核心任务, (锁定位置进行抓车和放车, 最高任务, 不允许打断和被打断)



private:

};


#endif //NNXX_TESTS_CARRIER_BASE_H
